/**
 * \file stars.c
 * \brief Starfield Effect Demo
 *
 * lionel_debroux@yahoo.fr (TICT) 2010-2011
 *
 * Port to the Nspire (Ndless pre-2.0) of the TI-68k version made in 2000-2004
 * by Thomas Nussbaumer & Lionel Debroux, with a modification: this version
 * uses a LFSR (pseudo-RNG) for computing the stars' position, instead of
 * a larger and slower RNG.
 *
 * This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details.
 */

#include <os.h>
#include <libndls.h>
#include <stdint.h>


#define WAIT(n, name) \
  asm volatile ( \
    "ldr r0, ="n"\n" \
    "_loop_"name":\n" \
    "subs r0, r0, #1\n" \
    "bne _loop_"name"\n" \
    : : : "r0" );


//! Generator for 32-bit numbers.
#define LFSR_XOR_MASK (0xA3000000)

#define BITS_PER_PIXEL  (4)
#define BITS_PER_BYTE   (8)
#define SCREEN_BYTES_PER_LINE (SCREEN_WIDTH * BITS_PER_PIXEL / BITS_PER_BYTE)


/*---------------------------------------------------------------------------*/
/* some constants defining the look of the starfield                         */
/*---------------------------------------------------------------------------*/
#define XOFFSET      (SCREEN_WIDTH>>1)
#define YOFFSET      (SCREEN_HEIGHT>>1)
#define ZOFFSET      254
#define MAX_STAR_X   (65535)
#define MAX_STAR_Y   (65535)
#define MAX_STARX2   (MAX_STAR_X>>1)
#define MAX_STARY2   (MAX_STAR_Y>>1)
#define BORDER_BLACK 80
#define BORDER_DGRAY 140
#define BORDER_LGRAY 190


//! Maximum number of stars on screen
#define MAX_STARS_COUNT        1000
//! Minimum number of stars on screen
#define MIN_STARS_COUNT        200
//! Number of stars changed by each keypress
#define STARS_COUNT_INCREMENT  50

//! Maximum "speed" in forward direction
#define MAX_ZDECREMENT            9
//! Maximum "speed" in backward direction
#define MIN_ZDECREMENT           -9


/*---------------------------------------------------------------------------*/
/* structure which stores a star's location                                  */
/*---------------------------------------------------------------------------*/
typedef struct {
    uint16_t offset;  ///< offset of old location for cleanup
    int16_t x;
    int16_t y;
    uint8_t z;
    uint8_t mask;    ///< inverted mask of old location for cleanup
} star_type;


star_type* stars = NULL;  // array of stars


static uint32_t stars_count;
static int8_t zdecrement;

//! Holds the LFSR value
uint32_t seq;


//! Initializes star with given index
static inline void CreateStar(star_type *ptr) {
    ptr->x = (seq >> 16) - 0x8000;
    ptr->y = (seq & 0xFFFF) - 0x8000;

    if (seq & 1) {
        seq = (seq >> 1) ^ LFSR_XOR_MASK;
    }
    else {
        seq = seq >> 1;
    }

    if (zdecrement > 0) {
        ptr->z = ZOFFSET;
    }
    else {
        ptr->z = ((seq >> 12) & 0xFE) + 1;
    }
}


//! Delete stars at old position, calculate new position, draw new position
static void HandleStars(void) {
    uint16_t      offset;
    uint8_t       mask;
    int16_t       drawx;
    int16_t       drawy;
    uint32_t      i;
    star_type     *ptr;

    ptr = &stars[0];

    for (i = 0; i < stars_count; i++) {
        if (ptr->offset != 0xFFFF) {
            *((uint8_t *)SCREEN_BASE_ADDRESS + ptr->offset) |= ptr->mask;
            ptr->offset = 0xFFFF;
        }

        // calculate new drawing x position
        // and check if it out of screen
        drawx = ptr->x / ptr->z + XOFFSET;
        //drawx = (int8_t)(seq >> 8) + XOFFSET;//ptr->x / ptr->z + XOFFSET;
        if (drawx < 0 || drawx >= SCREEN_WIDTH) {
            goto _CREATESTAR_;
        }

        // calculate new drawing y position
        // and check if it out of screen
        drawy = ptr->y / ptr->z + YOFFSET;
        //drawy = (int8_t)(seq >> 24) + XOFFSET;//ptr->y / ptr->z + YOFFSET;
        if (drawy < 0 || drawy >= SCREEN_HEIGHT) {
            goto _CREATESTAR_;
        }

        // calculate byte offset of pixel
        offset = (drawy * SCREEN_BYTES_PER_LINE) + (drawx * BITS_PER_PIXEL / BITS_PER_BYTE);

        // calculate mask
        mask = ((drawx & 1) ? 0x0F : 0xF0);

        // set pixel depending on z value
        if (ptr->z < BORDER_BLACK) {
            *((uint8_t *)SCREEN_BASE_ADDRESS + offset) &= ~(mask);
        }
        else if (ptr->z < BORDER_DGRAY) {
            *((uint8_t *)SCREEN_BASE_ADDRESS + offset) &= ~(mask & 0x44);
        }
        else if (ptr->z < BORDER_LGRAY) {
            *((uint8_t *)SCREEN_BASE_ADDRESS + offset) &= ~(mask & 0x88);
        }
        else {
            *((uint8_t *)SCREEN_BASE_ADDRESS + offset) &= ~(mask & 0xCC);
        }

        // store offset and bit number so that we can remove the pixel easily
        ptr->mask = mask;
        ptr->offset = offset;

        // decrement z and check if we must generate a new star
        ptr->z -= zdecrement;
        if (ptr->z <= 0 || ptr->z > ZOFFSET) {
_CREATESTAR_:
            CreateStar(ptr);
        }

        ptr++;
    }
}

//! Number of stars changed -> clear the screen
static inline void NrStarsChanged(void) {
    memset(SCREEN_BASE_ADDRESS, 0xFF, SCREEN_BYTES_SIZE);
}

//! Main routine: where all the fun starts...
int main(void) {
    uint32_t initial_contrast = *((uint32_t *)0x900F0020);

    // Reinitialize global variables
    seq = 1234567891;
    stars_count = 500;
    zdecrement = 4;

    show_msgbox(
        "star - Starfield effect demo",
        "------------------------\n"
        "  Starfield effect demo\n"
        "    © 2000-2010 TICT\n"
        "------------------------\n"
        "UP    - Increment number of stars\n"
        "DOWN  - Decrement number of stars\n"
        "RIGHT - Increase contrast\n"
        "LEFT  - Decrease contrast\n"
        ")     - Increase speed\n"
        "(     - Decrease speed\n"
        "ESC   - Exit"
    );

    stars = malloc(sizeof(star_type)*MAX_STARS_COUNT);
    if (stars != NULL) {

        // Clear screen
        memset(SCREEN_BASE_ADDRESS, 0xFF, SCREEN_BYTES_SIZE);

        // Setup initial star positions
        {
            uint32_t i;
            star_type *ptr = &stars[0];
            for (i = 0; i < MAX_STARS_COUNT; i++) {
                CreateStar(ptr);
                ptr->offset = -1;
                ptr++;
            }
        }

        // Main loop
        while (!isKeyPressed(KEY_NSPIRE_ESC)) {
            HandleStars();
            WAIT("0x1FFFFF", "loop_slowdown");
            if (isKeyPressed(KEY_NSPIRE_UP) && (stars_count <= (MAX_STARS_COUNT - STARS_COUNT_INCREMENT))) {
                stars_count += STARS_COUNT_INCREMENT;
                NrStarsChanged();
            }
            else if (isKeyPressed(KEY_NSPIRE_DOWN) && (stars_count >= (MIN_STARS_COUNT + STARS_COUNT_INCREMENT))) {
                stars_count -= STARS_COUNT_INCREMENT;
                NrStarsChanged();
            }
            else if (isKeyPressed(KEY_NSPIRE_RIGHT) && (*((uint32_t *)0x900F0020) < 0x95)) {
                // Increase contrast.
                *(uint32_t *)0x900F0020 += 1;
            }
            else if (isKeyPressed(KEY_NSPIRE_LEFT) && (*((uint32_t *)0x900F0020) > 0x6B)) {
                // Decrease contrast.
                *(uint32_t *)0x900F0020 -= 1;
            }
            else if (isKeyPressed(KEY_NSPIRE_LP) && (zdecrement>MIN_ZDECREMENT)) {
                zdecrement--;
                if (!zdecrement) zdecrement--;
            }
            else if (isKeyPressed(KEY_NSPIRE_RP) && (zdecrement<MAX_ZDECREMENT)) {
                zdecrement++;
                if (!zdecrement) zdecrement++;
            }
        }

        // Cleanup and bye =8-)
        // Restore contrast
        *((uint32_t *)0x900F0020) = initial_contrast;
        free(stars);
    }

    return 0;
}
